require 'zip_search'

module ZipSearch
  module HasLocations
    extend ActiveSupport::Concern

    module ClassMethods
      def has_location(title=:location, options={})
        has_one title, -> { where zs_association_name: title.to_s },
          as: :locatable, class_name: 'ZipSearch::Location'
        initialize_location_ownership! title
      end

      def has_locations(title=:locations, options={})
        has_many title, -> { where zs_association_name: title.to_s },
          as: :locatable, class_name: 'ZipSearch::Location'
        initialize_locations_ownership! title
      end

      def has_location_history(title=:location_history)
        has_locations title
        has_location :current_location
        before_validation if: -> { current_location && current_location.changed? } do
          historical_attrs = {}
          current_location.changes.each{|k, (oldval, _)|
            historical_attrs[k] = oldval unless oldval.blank? }
          if historical_attrs.any?
            self.location_history << Location.new( Hash[
              current_location.changes.map{|k, (v, _)| [ k, v ] } ].merge(
                zs_association_name: 'location_history') )
          end
        end
      end

      def zipsearch_associations
        reflect_on_all_associations.select{|a| a.foreign_key == 'locatable_id' }
      end
      def zipsearch_association_names
        zipsearch_associations.map(&:name) end
      def zipsearch_association(title=nil)
        zipsearch_associations.find{|za| title.nil? || za.name == title } end

      def has_zipsearch_association?(title=nil)
        return zipsearch_associations.any? if title.nil?
        zipsearch_association_names.include?(title)
      end

      def initialize_location_ownership!(title=:location)
        include HasLocations::LocalInstanceMethods
        accepts_nested_attributes_for title, allow_destroy: true,
          reject_if: :"should_reject_#{title}?"
        before_validation :prune_empty_zs_locations
        after_initialize if: :"should_build_#{title}?" do
          send :"build_#{title}", zs_association_name: title end
      end
      def initialize_locations_ownership!(title=:locations)
        include HasLocations::LocalInstanceMethods
        accepts_nested_attributes_for title, allow_destroy: true,
          reject_if: :"should_reject_#{title}?"
        before_validation :prune_empty_zs_locations
        unless title == :location_history
          after_initialize if: :"should_build_#{title}?" do
            send(title).build zs_association_name: title end
        end
      end
    end

    module LocalInstanceMethods

      ZipSearch::ASSOCIATION_MODES.each do |am|
        define_method :"should_build_#{am}?" do
          respond_to?(am) && send(am).nil? end
        define_method :"should_reject_#{am}?" do |attributed|
          self.class.zipsearch_association(am).klass
              .location_fields.select{|lf| !attributed[lf].blank? }.empty? end
      end

      def prune_empty_zs_locations
        self.class.zipsearch_associations.each{|zsa|
              Array(self.send(zsa.name)).each {|zsl|
                zsl.delete if zsl.new_record? && zsl.location_blank? } }
      end

    end

  end
end
